/*******************************************************************
 * 
 * Input switcher for Headphone amplifier
 * 2 status LED. 
 *    1th color for selected input 
 *    2th red=power on (output relay inactive), green=ready (output relay active)
 * Mute Output 
 * Ready input
 * 
 * Andreas Brietzke, Nagareyama, J
 * 2016/01/17
 *
 * Chip type           : ATtiny2313, Tiny13, Tiny85
 * Clock frequency     : 1 MHz
 * Fuses H:D9, L:E4 Tiny2313
 * Fuses H:DC, L:E2 Tiny85
 * 
 *     AudioSwitcher.c
 * 
 *     key routines from Peter Dannegger
 * 
 *    Using: 
 *    press key short: Switch to the next input
 *    press key long: Switch to the previous input
 *    press key very long: Switch to SET MODE
 *       press key short in SET MODE: Increase the count of audio inputs. If MAX_INPUT is reach then switch to 1 again
 *       press key long in SET MODE: Leave SET MODE and store the settings
 *       no action for 10s in SET MODE: Leave SET MODE store the settings
 * 
 *    If defined inputs less than MAX_INPUT then the last input will parallel switched to the selected input. (in this case will use this input as output for recording)
 * 
 ********************************************************************/
#include "AudioSwitcher.h"
#include <util/atomic.h>

#define REL1_PIN        PD2  // purple
#define REL2_PIN        PD5  // blue
#define REL3_PIN        PD4  // green
#define REL4_PIN        PD3  // red
#define REL5_PIN        PD1
#define REL6_PIN        PD0
#define PORT_REL        PORTD
#define DDR_REL         DDRD

#define REL1A_PIN       PA0
#define REL2A_PIN       PA1
#define PORTA_REL       PORTA
#define DDRA_REL        DDRA

#define KEY_DDR         DDRB
#define KEY_PORT        PORTB
#define KEY_PIN         PINB
#define KEY0            PB4   // brown, switch to GND
#define ACTIVE          PB2   // purple, actice = 1 Input
#define ALL_KEYS        (1<<KEY0)

#define DDR_MUTE        DDRB
#define MUTE            PB1   // white, active = 1 Output
#define MUTE_ON         PORTB |= 1<<PB1
#define MUTE_OFF        PORTB &= ~(1<<PB1)

#define REPEAT_MASK     (1<<KEY0)
#define REPEAT_START    30    // after 300ms
#define REPEAT_NEXT     250   // every 2500ms

#define MAX_INPUT       4     // physical inputs,  adapt "InitPortArray"!

#ifndef F_CPU
   #define F_CPU 1000000UL
#endif
#define UART_BAUD_RATE 38400
#define UART_BAUD_CALC(UART_BAUD_RATE,F_CPU) ((F_CPU)/((UART_BAUD_RATE)*16)-1)



uint8_t Port[MAX_INPUT];      // adapt "InitPortArray"!

uint8_t key_state;            // debounced and inverted key state: bit = 1: key pressed
uint8_t key_press;            // key press detect
uint8_t key_rpt;              // key long press and repeat

uint8_t DummyByte EEMEM;      // sometimes 1nd adress can't use
uint8_t eePortStatus EEMEM; 
uint8_t eeMaxInput EEMEM;

struct cRGB led[2];

#if ((__AVR_ATtiny2313__) || (__AVR_ATtiny85__))
ISR(TIMER0_OVF_vect) {        // every 10ms attiny85 tiny2313
#elif defined( __AVR_ATtiny13__ )
ISR(TIM0_OVF_vect) {        // every 10ms attiny13
#else
#  error "Unknown MCU type."
#endif  

   
  static uint8_t ct0, ct1, rpt;
  uint8_t i;
  
  TCNT0 = (uint8_t)(int16_t)-(F_CPU / 1024 * 10e-3 + 0.5);  // preload for 10ms
  
  i = key_state ^ ~KEY_PIN;                           // key changed ?
  ct0 = ~( ct0 & i );                                 // reset or count ct0
  ct1 = ct0 ^ (ct1 & i);                              // reset or count ct1
  i &= ct0 & ct1;                                     // count until roll over ?
  key_state ^= i;                                     // then toggle debounced state
  key_press |= key_state & i;                         // 0->1: key press detect
 
  if( (key_state & REPEAT_MASK) == 0 )                // check repeat function
     rpt = REPEAT_START;                              // start delay
  if( --rpt == 0 ){
    rpt = REPEAT_NEXT;                                // repeat delay
    key_rpt |= key_state & REPEAT_MASK;
  }
}

// Allowed functions of a key:
// - get_key_press()
// - get_key_rpt()
// - get_key_press() mit get_key_rpt()
// - get_key_short() mit get_key_long()
// - get_key_short() mit get_key_long_r() und get_key_rpt_l()

uint8_t get_key_press( uint8_t key_mask ) {
   ATOMIC_BLOCK(ATOMIC_FORCEON){   
      key_mask &= key_press;                          // read key(s)
      key_press ^= key_mask;                          // clear key(s)
   }
   return key_mask;
}

uint8_t get_key_rpt( uint8_t key_mask ) {
   ATOMIC_BLOCK(ATOMIC_FORCEON){   
      key_mask &= key_rpt;                            // read key(s)
      key_rpt ^= key_mask;                            // clear key(s)
   }
   return key_mask;
}

uint8_t get_key_short( uint8_t key_mask )
{
  uint8_t i;

  ATOMIC_BLOCK(ATOMIC_FORCEON)
    i = get_key_press( ~key_state & key_mask );
  return i;
}

uint8_t get_key_long( uint8_t key_mask ) {
   return get_key_press( get_key_rpt( key_mask ));
}

uint8_t get_key_long_r( uint8_t key_mask )            // if repeat function needed
{
   return get_key_press( get_key_rpt( key_press & key_mask ));
}

uint8_t get_key_rpt_l( uint8_t key_mask )             // if long function needed
{
   return get_key_rpt( ~key_press & key_mask );
}

void SwitchLed (uint8_t stat) {
   if (stat == Port[0]) {
      led[0].r=40;led[0].g=0;led[0].b=40;             // channel 1: violet
      ws2812_setleds(led,1);
   } else if (stat == Port[1]) {
      led[0].r=0;led[0].g=0;led[0].b=60;              // channel 2: blue
      ws2812_setleds(led,1);
   } else if (stat == Port[2]) {
      led[0].r=0;led[0].g=40;led[0].b=0;              // channel 3: green
      ws2812_setleds(led,1);
   } else if (stat == Port[3]) {
      led[0].r=60;led[0].g=0;led[0].b=0;              // channel 4: red
      ws2812_setleds(led,1);
   }  else {
      led[0].r=40;led[0].g=40;led[0].b=40;            // channel 4: white, invalid!
      ws2812_setleds(led,1);      
   }
}            

void StatusRed(void) {
   led[1].r=50;led[1].g=0;led[1].b=0;
   ws2812_setleds(led,2);
}

void StatusGreen(void) {
   led[1].r=0;led[1].g=50;led[1].b=0;
   ws2812_setleds(led,2);
}

void SwitchInput(uint8_t PortStatus, uint8_t PortStatusOld) {
   MUTE_ON;
   StatusRed();
   _delay_ms(50);
   PORT_REL = 0;
   _delay_ms(100);
   PORT_REL = PortStatus;
   PortStatusOld = PortStatus;
   eeprom_write_byte(&eePortStatus, PortStatus);
   _delay_ms(100);
   MUTE_OFF;
   SwitchLed(PortStatus);   
}

void InitPortArray(uint8_t MaxInput) {
   if (MaxInput < MAX_INPUT){                         // last input is switched parallel to the selected input
      Port[3] = 1<<REL4_PIN;
      Port[0] = 1<<REL1_PIN | Port[MAX_INPUT -1];
      Port[1] = 1<<REL2_PIN | Port[MAX_INPUT -1];
      Port[2] = 1<<REL3_PIN | Port[MAX_INPUT -1];
   } else {
      Port[0] = 1<<REL1_PIN;
      Port[1] = 1<<REL2_PIN;
      Port[2] = 1<<REL3_PIN;
      Port[3] = 1<<REL4_PIN;   
   }
}

int main( void ) {
   uint8_t PortStatus;                                // status of the switch port
   uint8_t PortStatusOld;
   uint8_t i;
   uint8_t setmode = 0;
   uint8_t MaxInput = MAX_INPUT;
   uint8_t Timer = 0;
   
   // Outputs
   DDR_REL |= (1<<REL1_PIN | 1<<REL2_PIN | 1<<REL3_PIN | 1<<REL4_PIN| 1<<REL5_PIN | 1<<REL6_PIN);
   DDR_MUTE |= 1<<MUTE;
   DDRA_REL |= (1<<REL1A_PIN | 1<<REL2A_PIN);         // reserve, connected to ULN2803
   KEY_PORT |= (1 << KEY0);                           // enable pull-up resistor

   // init timer 1
   TCCR0B = 1<<CS02 | 1<<CS00;                        // divide by 1024
   
#if ((__AVR_ATtiny2313__) || (__AVR_ATtiny85__))
   TIMSK = 1<<TOIE0;                                  // enable timer 0 overflow interrupt. attiny85, tiny 2313
   TCNT0 = (uint8_t)(int16_t)-(F_CPU / 1024 * 10e-3 + 0.5);  // preload for 10ms   
   TIFR = 1<<TOV0;                                    // T0 OVF IRQ. attiny85, 2313
#elif defined( __AVR_ATtiny13__ )
   TIMSK0 = 1<<TOIE0;                                 // enable timer 0 overflow interrupt. attiny13
   TCNT0 = (uint8_t)(int16_t)-(F_CPU / 1024 * 10e-3 + 0.5);  // preload for 10ms   
   TIFR0 = 1<<TOV0;                                   // T0 OVF IRQ. attiny13
#else
#  error "Unknown MCU type."
#endif  

   MaxInput = eeprom_read_byte(&eeMaxInput);
   if (MaxInput > MAX_INPUT) {                        // unvalid value
      MaxInput = MAX_INPUT;
      eeprom_write_byte(&eeMaxInput,MaxInput);
   }

   InitPortArray(MaxInput);
   
   PortStatus = eeprom_read_byte(&eePortStatus);
   for (i=1; i<MaxInput; i++) {
      if (PortStatus==Port[i]) {
         i=255;
         break;
      }
   }
   if (i!=255) {                                      // no valid value in eeprom
      PortStatus=Port[0];                    
      eeprom_write_byte(&eePortStatus, PortStatus);
   }
   
   StatusRed();
   _delay_ms(300);
   StatusGreen();
   _delay_ms(300);
   StatusRed();
   _delay_ms(300);
   StatusGreen();
   _delay_ms(300);
   StatusRed();
 
   for (i=0; i < MaxInput; i++) {
      SwitchLed(Port[i]);
      _delay_ms(300);
   }   
   
   PORT_REL = PortStatus;
   PortStatusOld = PortStatus;
   SwitchLed(PortStatus);   
   
   while (1) {
      if (setmode) {
         for (i=0; i < MaxInput; i++) {
            led[0].r=40;led[0].g=40;led[0].b=40;
            ws2812_setleds(led,1);
            _delay_ms(5);
            led[0].r=0;led[0].g=0;led[0].b=0;
            ws2812_setleds(led,1);
            _delay_ms(150);
         }
         _delay_ms(300);
         
         if(get_key_short(1<<KEY0)){
            MaxInput++;
            if (MaxInput > MAX_INPUT) MaxInput=1;
            InitPortArray(MaxInput);
            Timer=10;
         }
         if(get_key_long_r(1<<KEY0)){
            Timer=0;
         }
         if (!(Timer--)) {                            // end set mode
            setmode=0;
            eeprom_write_byte(&eeMaxInput,MaxInput);
            PortStatus=Port[0];
            SwitchInput(PortStatus, PortStatusOld);
         }
      } else {                                        // no set mode
         
         if(get_key_rpt_l(1<<KEY0)){                  // mode
            setmode=1;
            Timer=10;
         }
         
         if(get_key_long_r(1<<KEY0)){                 // switch backward
            if (PortStatus==Port[0]) {
               PortStatus=Port[MaxInput-1];
            } else {
               for (i=1; i<MaxInput; i++) {
                  if (PortStatus==Port[i]) {
                     PortStatus=Port[i-1];
                     break;
                  }
               }                     
            }
            SwitchInput(PortStatus, PortStatusOld);
         }
         
         if(get_key_short(1<<KEY0)){                  // switch forward
            if (PortStatus==Port[MaxInput-1]) {
               PortStatus=Port[0];
            } else {
               for (i=0; i<MaxInput-1; i++) {
                  if (PortStatus==Port[i]) {
                     PortStatus=Port[i+1];
                     break;
                  }
               }                     
            }
            SwitchInput(PortStatus, PortStatusOld);
         }
         
         if (KEY_PIN & 1<<ACTIVE) StatusRed(); else StatusGreen();
         
      }
   }
}

